//+------------------------------------------------------------------+
//|                                                     template.mq5 |
//|                                  Copyright 2022, MetaQuotes Ltd. |
//|                                             https://www.mql5.com |
//+------------------------------------------------------------------+
#property copyright "Copyright 2022, MetaQuotes Ltd."
#property link      "https://www.mql5.com"
#property version   "1.00"

#include             <ct_19.mqh>
#property            script_show_inputs
//+------------------------------------------------------------------+
//| INPUTS                                                           |
//+------------------------------------------------------------------+
input string         __currency_a="USDJPY.ln";
input string         __currency_b="EURJPY.ln";
input string         __currency_c="EURUSD.ln";

input datetime       __training_start=D'2016.08.01';
input datetime       __training_stop=D'2022.08.01';
input int            __training_trees=75;
input double         __training_r=0.382;
input datetime       __testing_stop=D'2023.08.01';
input int            __n_inductions=1;
//+------------------------------------------------------------------+
//| Script program start function                                    |
//+------------------------------------------------------------------+
void OnStart()
   {
      //create domain series (USDJPY-rates) in first category for training
      CCategory _c_a_train;_c_a_train.Let();
      int _a_size=GetObject(_c_a_train,__currency_a,__training_start,__training_stop);
      
      //create 2 series (EURJPY and EURUSD rates) in second category for training
      CCategory _c_bc_train;_c_bc_train.Let();
      int _b_trains=GetObject(_c_bc_train,__currency_b,__training_start,__training_stop);
      int _c_trains=GetObject(_c_bc_train,__currency_c,__training_start,__training_stop);
      
      //create natural transformation, by induction across, n squares..., cpi to pmi
      //mapping by random forests
      int _training_size=fmin(_c_trains,_b_trains);
      int _info_ab=0,_info_bd=0,_info_ac=0,_info_cd=0;
      CDFReport _report_ab,_report_bd,_report_ac,_report_cd;
      CMatrixDouble _xy_ab;_xy_ab.Resize(_training_size,1+1);
      CMatrixDouble _xy_bd;_xy_bd.Resize(_training_size,1+1);
      CMatrixDouble _xy_ac;_xy_ac.Resize(_training_size,1+1);
      CMatrixDouble _xy_cd;_xy_cd.Resize(_training_size,1+1);
      
      double _a=0.0,_b=0.0,_c=0.0,_d=0.0;
      CElement<string> _e_a,_e_b,_e_c,_e_d;
      string _s_a="",_s_b="",_s_c="",_s_d="";
      for(int i=0;i<_training_size-__n_inductions;i++)
      {
         _s_a="";_e_a.Let();_c_bc_train.domain[i].Get(0,_e_a);_e_a.Get(1,_s_a);_a=StringToDouble(_s_a);
         _s_b="";_e_b.Let();_c_bc_train.domain[i+_b_trains].Get(0,_e_b);_e_b.Get(1,_s_b);_b=StringToDouble(_s_b);
         _s_c="";_e_c.Let();_c_bc_train.domain[i+__n_inductions].Get(0,_e_c);_e_c.Get(1,_s_c);_c=StringToDouble(_s_c);
         _s_d="";_e_d.Let();_c_bc_train.domain[i+_b_trains+__n_inductions].Get(0,_e_d);_e_d.Get(1,_s_d);_d=StringToDouble(_s_d);
         
         if(i<_training_size-__n_inductions)
         {
            _xy_ab[i].Set(0,_a);
            _xy_ab[i].Set(1,_b);
            
            _xy_bd[i].Set(0,_b);
            _xy_bd[i].Set(1,_d);
            
            _xy_ac[i].Set(0,_a);
            _xy_ac[i].Set(1,_c);
            
            _xy_cd[i].Set(0,_c);
            _xy_cd[i].Set(1,_d);
         }
      }
      
      CDForest          _forest;
      CDecisionForest   _rdf_ab,_rdf_cd;
      CDecisionForest   _rdf_ac,_rdf_bd;
      
      _forest.DFBuildRandomDecisionForest(_xy_ab,_training_size-__n_inductions,1,1,__training_trees,__training_r,_info_ab,_rdf_ab,_report_ab);
      _forest.DFBuildRandomDecisionForest(_xy_bd,_training_size-__n_inductions,1,1,__training_trees,__training_r,_info_bd,_rdf_bd,_report_bd);
      _forest.DFBuildRandomDecisionForest(_xy_ac,_training_size-__n_inductions,1,1,__training_trees,__training_r,_info_ac,_rdf_ac,_report_ac);
      _forest.DFBuildRandomDecisionForest(_xy_cd,_training_size-__n_inductions,1,1,__training_trees,__training_r,_info_cd,_rdf_cd,_report_cd);
      
      //
      if(_info_ab>0 && _info_bd>0 && _info_ac>0 && _info_cd>0)
      {
         //create 2 objects (cpi and pmi) in second category for testing
         CCategory _c_cp_test;_c_cp_test.Let();
         int _b_test=GetObject(_c_cp_test,__currency_b,__training_stop,__testing_stop);
         int _c_test=GetObject(_c_cp_test,__currency_c,__training_stop,__testing_stop);
      
         int _strict_match=0,_generic_match=0,_miss=0;
         int _testing_size=fmin(_b_test,_c_test);
         
         MqlRates _rates[];
         ArraySetAsSeries(_rates,true);
         if(CopyRates(__currency_c,Period(), 0, _testing_size+__n_inductions+1, _rates)>=_testing_size+__n_inductions+1)
         {
            ArraySetAsSeries(_rates,true);
            
            for(int i=__n_inductions+_testing_size;i>__n_inductions;i--)
            {
               _s_a="";_e_a.Let();_c_cp_test.domain[i].Get(0,_e_a);_e_a.Get(1,_s_a);_a=StringToDouble(_s_a);
               
               double _x_ab[],_y_ab[]; ArrayResize(_x_ab,1); ArrayResize(_y_ab,1);
               ArrayInitialize(_x_ab,0.0); ArrayInitialize(_y_ab,0.0);
               //
               _x_ab[0]=_a; _forest.DFProcess(_rdf_ab,_x_ab,_y_ab);
               
               double _x_bd[],_y_bd[]; ArrayResize(_x_bd,1); ArrayResize(_y_bd,1);
               ArrayInitialize(_x_bd,0.0); ArrayInitialize(_y_bd,0.0);
               //
               _x_bd[0]=_y_ab[0]; _forest.DFProcess(_rdf_bd,_x_bd,_y_bd);
               
               double _x_ac[],_y_ac[]; ArrayResize(_x_ac,1); ArrayResize(_y_ac,1);
               ArrayInitialize(_x_ac,0.0); ArrayInitialize(_y_ac,0.0);
               //
               _x_ac[0]=_a; _forest.DFProcess(_rdf_ac,_x_ac,_y_ac);
               
               double _x_cd[],_y_cd[]; ArrayResize(_x_cd,1); ArrayResize(_y_cd,1);
               ArrayInitialize(_x_cd,0.0); ArrayInitialize(_y_cd,0.0);
               //
               _x_cd[0]=_y_ac[0]; _forest.DFProcess(_rdf_cd,_x_cd,_y_cd);
               
               double _c_forecast=0.0;
               if((_y_bd[0]>0.0 && _y_cd[0]>0.0)||(_y_bd[0]<0.0 && _y_cd[0]<0.0))//abd agrees with acd on currency c change
               {
                  _c_forecast=0.5*(_y_bd[0]+_y_cd[0]);
               }
               
               double _c_actual=_rates[i-__n_inductions].close-_rates[i].close;
               
               if((_c_forecast>0.0 && _c_actual>0.0)||(_c_forecast<0.0 && _c_actual<0.0)){ _strict_match++; }
               else if((_c_forecast>=0.0 && _c_actual>=0.0)||(_c_forecast<=0.0 && _c_actual<=0.0)){ _generic_match++; }
               else { _miss++; }
            }
            //
            printf(__FUNCSIG__+" misses: "+IntegerToString(_miss)+
                        ", strict matches: "+IntegerToString(_strict_match)+
                        ", & generic matches: "+IntegerToString(_generic_match)+
                        ", for strict pct (excl. generic): "+DoubleToString(double(_strict_match)/(double(_strict_match+_miss)),2)+
                        ", & generic pct: "+DoubleToString(double(_strict_match+_generic_match)/double(_strict_match+_generic_match+_miss),2)+
                        ", with inductions at: "+IntegerToString(__n_inductions)
                        );
         }
      }
   } 
//+------------------------------------------------------------------+
//| Custom Method to fill object by Type                             |
//+------------------------------------------------------------------+
int GetObject(CCategory &C,string Currency,datetime Start,datetime Stop)
   {
      CDomain<string> _d;
      CElement<string> _e;
      
      int _domains=C.Domains();
      if(_domains<0){ _domains=0; }
      
      MqlRates _rates[];
      ArraySetAsSeries(_rates,true);
      int _rates_copied=CopyRates(Currency,Period(), Start, Stop, _rates);
      ArraySetAsSeries(_rates,true);
      //printf(__FUNCSIG__+" for currency: "+Currency+" rates got are: "+IntegerToString(_rates_copied)+", while objects are: "+IntegerToString(_domains));
      //---
      for(int v=0;v<_rates_copied-1;v++)
      {
         GetElement(C,_d,_e,_rates[v],_rates[v+1]);
      }
      
      return(C.Domains()-_domains);
   }
//+------------------------------------------------------------------+
//| Custom Method to fill element by Type                            |
//+------------------------------------------------------------------+
void GetElement(CCategory &C,CDomain<string> &D,CElement<string> &E,MqlRates &New,MqlRates &Old)
   {
      C.Domains(C.Domains()+1);
      D.Let();D.Cardinality(1);
      E.Let();E.Cardinality(2);
      E.Set(0,TimeToString(New.time));
      E.Set(1,DoubleToString(New.close-Old.close,2));
      D.Set(0,E);
      C.SetDomain(C.Domains()-1,D);
   }
//+------------------------------------------------------------------+
